@@ -5,16 +5,16 @@ use crate::{
     RefChange, RefChangeChannel, SourceInfo, SourceMeta,
 };
 use anyhow::Result;
-use bevy_ecs::system::Res;
+use bevy_ecs::system::{Res, ResMut};
 use bevy_log::warn;
 use bevy_tasks::TaskPool;
 use bevy_utils::{HashMap, Uuid};
 use crossbeam_channel::TryRecvError;
-use parking_lot::RwLock;
+use parking_lot::{Mutex, RwLock};
 use std::{collections::hash_map::Entry, path::Path, sync::Arc};
 use thiserror::Error;
 
-/// Errors that occur while loading assets with an AssetServer
+/// Errors that occur while loading assets with an `AssetServer`
 #[derive(Error, Debug)]
 pub enum AssetServerError {
     #[error("asset folder path is not a directory: {0}")]
@@ -45,6 +45,7 @@ fn format_missing_asset_ext(exts: &[String]) -> String {
 pub(crate) struct AssetRefCounter {
     pub(crate) channel: Arc<RefChangeChannel>,
     pub(crate) ref_counts: Arc<RwLock<HashMap<HandleId, usize>>>,
+    pub(crate) mark_unused_assets: Arc<Mutex<Vec<HandleId>>>,
 }
 
 pub struct AssetServerInternal {
@@ -52,7 +53,7 @@ pub struct AssetServerInternal {
     pub(crate) asset_ref_counter: AssetRefCounter,
     pub(crate) asset_sources: Arc<RwLock<HashMap<SourcePathId, SourceInfo>>>,
     pub(crate) asset_lifecycles: Arc<RwLock<HashMap<Uuid, Box<dyn AssetLifecycle>>>>,
-    loaders: RwLock<Vec<Arc<Box<dyn AssetLoader>>>>,
+    loaders: RwLock<Vec<Arc<dyn AssetLoader>>>,
     extension_to_loader_index: RwLock<HashMap<String, usize>>,
     handle_to_path: Arc<RwLock<HashMap<HandleId, AssetPath<'static>>>>,
     task_pool: TaskPool,
@@ -111,7 +112,7 @@ impl AssetServer {
                 .write()
                 .insert(extension.to_string(), loader_index);
         }
-        loaders.push(Arc::new(Box::new(loader)));
+        loaders.push(Arc::new(loader));
     }
 
     pub fn watch_for_changes(&self) -> Result<(), AssetServerError> {
@@ -129,15 +130,14 @@ impl AssetServer {
         HandleUntyped::strong(id.into(), sender)
     }
 
-    fn get_asset_loader(
-        &self,
-        extension: &str,
-    ) -> Result<Arc<Box<dyn AssetLoader>>, AssetServerError> {
-        self.server
-            .extension_to_loader_index
-            .read()
-            .get(extension)
-            .map(|index| self.server.loaders.read()[*index].clone())
+    fn get_asset_loader(&self, extension: &str) -> Result<Arc<dyn AssetLoader>, AssetServerError> {
+        let index = {
+            // scope map to drop lock as soon as possible
+            let map = self.server.extension_to_loader_index.read();
+            map.get(extension).copied()
+        };
+        index
+            .map(|index| self.server.loaders.read()[index].clone())
             .ok_or_else(|| AssetServerError::MissingAssetLoader {
                 extensions: vec![extension.to_string()],
             })
@@ -146,7 +146,7 @@ impl AssetServer {
     fn get_path_asset_loader<P: AsRef<Path>>(
         &self,
         path: P,
-    ) -> Result<Arc<Box<dyn AssetLoader>>, AssetServerError> {
+    ) -> Result<Arc<dyn AssetLoader>, AssetServerError> {
         let s = path
             .as_ref()
             .file_name()
@@ -204,6 +204,7 @@ impl AssetServer {
                     }
                     LoadState::Failed => return LoadState::Failed,
                     LoadState::NotLoaded => return LoadState::NotLoaded,
+                    LoadState::Unloaded => return LoadState::Unloaded,
                 },
                 HandleId::Id(_, _) => return LoadState::NotLoaded,
             }
@@ -212,18 +213,33 @@ impl AssetServer {
         load_state
     }
 
+    /// Queue an [`Asset`] at the provided relative path for asynchronous loading.
+    ///
+    /// The absolute Path to the asset is `"ROOT/ASSET_FOLDER_NAME/path"`.
+    ///
+    /// By default the ROOT is the directory of the Application, but this can be overridden by
+    /// setting the `"CARGO_MANIFEST_DIR"` environment variable
+    /// (see <https://doc.rust-lang.org/cargo/reference/environment-variables.html>)
+    /// to another directory. When the application  is run through Cargo, then
+    /// `"CARGO_MANIFEST_DIR"` is automatically set to the root folder of your crate (workspace).
+    ///
+    /// The name of the asset folder is set inside the
+    /// [`AssetServerSettings`](crate::AssetServerSettings) resource. The default name is
+    /// `"assets"`.
+    ///
+    /// The asset is loaded asynchronously, and will generally not be available by the time
+    /// this calls returns. Use [`AssetServer::get_load_state`] to determine when the asset is
+    /// effectively loaded and available in the [`Assets`] collection.
+    #[must_use = "not using the returned strong handle may result in the unexpected release of the asset"]
     pub fn load<'a, T: Asset, P: Into<AssetPath<'a>>>(&self, path: P) -> Handle<T> {
         self.load_untyped(path).typed()
     }
 
-    // TODO: properly set failed LoadState in all failure cases
-    async fn load_async<'a, P: Into<AssetPath<'a>>>(
+    async fn load_async(
         &self,
-        path: P,
+        asset_path: AssetPath<'_>,
         force: bool,
     ) -> Result<AssetPathId, AssetServerError> {
-        let asset_path: AssetPath = path.into();
-        let asset_loader = self.get_path_asset_loader(asset_path.path())?;
         let asset_path_id: AssetPathId = asset_path.get_id();
 
         // load metadata and update source info. this is done in a scope to ensure we release the
@@ -242,11 +258,12 @@ impl AssetServer {
                 }),
             };
 
-            // if asset is already loaded (or is loading), don't load again
+            // if asset is already loaded or is loading, don't load again
             if !force
-                && source_info
+                && (source_info
                     .committed_assets
                     .contains(&asset_path_id.label_id())
+                    || source_info.load_state == LoadState::Loading)
             {
                 return Ok(asset_path_id);
             }
@@ -258,15 +275,28 @@ impl AssetServer {
             source_info.version
         };
 
+        let set_asset_failed = || {
+            let mut asset_sources = self.server.asset_sources.write();
+            let source_info = asset_sources
+                .get_mut(&asset_path_id.source_path_id())
+                .expect("`AssetSource` should exist at this point.");
+            source_info.load_state = LoadState::Failed;
+        };
+
+        // get the according asset loader
+        let asset_loader = match self.get_path_asset_loader(asset_path.path()) {
+            Ok(loader) => loader,
+            Err(err) => {
+                set_asset_failed();
+                return Err(err);
+            }
+        };
+
         // load the asset bytes
         let bytes = match self.server.asset_io.load_path(asset_path.path()).await {
             Ok(bytes) => bytes,
             Err(err) => {
-                let mut asset_sources = self.server.asset_sources.write();
-                let source_info = asset_sources
-                    .get_mut(&asset_path_id.source_path_id())
-                    .expect("`AssetSource` should exist at this point.");
-                source_info.load_state = LoadState::Failed;
+                set_asset_failed();
                 return Err(AssetServerError::AssetIoError(err));
             }
         };
@@ -277,11 +307,17 @@ impl AssetServer {
             &self.server.asset_ref_counter.channel,
             &*self.server.asset_io,
             version,
+            &self.server.task_pool,
         );
-        asset_loader
+
+        if let Err(err) = asset_loader
             .load(&bytes, &mut load_context)
             .await
-            .map_err(AssetServerError::AssetLoaderError)?;
+            .map_err(AssetServerError::AssetLoaderError)
+        {
+            set_asset_failed();
+            return Err(err);
+        }
 
         // if version has changed since we loaded and grabbed a lock, return. theres is a newer
         // version being loaded
@@ -313,7 +349,7 @@ impl AssetServer {
             let type_uuid = loaded_asset.value.as_ref().unwrap().type_uuid();
             source_info.asset_types.insert(label_id, type_uuid);
             for dependency in loaded_asset.dependencies.iter() {
-                self.load_untyped(dependency.clone());
+                self.load_untracked(dependency.clone(), false);
             }
         }
 
@@ -325,17 +361,13 @@ impl AssetServer {
         Ok(asset_path_id)
     }
 
+    #[must_use = "not using the returned strong handle may result in the unexpected release of the asset"]
     pub fn load_untyped<'a, P: Into<AssetPath<'a>>>(&self, path: P) -> HandleUntyped {
-        let handle_id = self.load_untracked(path, false);
+        let handle_id = self.load_untracked(path.into(), false);
         self.get_handle_untyped(handle_id)
     }
 
-    pub(crate) fn load_untracked<'a, P: Into<AssetPath<'a>>>(
-        &self,
-        path: P,
-        force: bool,
-    ) -> HandleId {
-        let asset_path: AssetPath<'a> = path.into();
+    pub(crate) fn load_untracked(&self, asset_path: AssetPath<'_>, force: bool) -> HandleId {
         let server = self.clone();
         let owned_path = asset_path.to_owned();
         self.server
@@ -346,9 +378,18 @@ impl AssetServer {
                 }
             })
             .detach();
+
+        let handle_id = asset_path.get_id().into();
+        self.server
+            .handle_to_path
+            .write()
+            .entry(handle_id)
+            .or_insert_with(|| asset_path.to_owned());
+
         asset_path.into()
     }
 
+    #[must_use = "not using the returned strong handles may result in the unexpected release of the assets"]
     pub fn load_folder<P: AsRef<Path>>(
         &self,
         path: P,
@@ -378,10 +419,35 @@ impl AssetServer {
     }
 
     pub fn free_unused_assets(&self) {
+        let mut potential_frees = self.server.asset_ref_counter.mark_unused_assets.lock();
+
+        if !potential_frees.is_empty() {
+            let ref_counts = self.server.asset_ref_counter.ref_counts.read();
+            let asset_sources = self.server.asset_sources.read();
+            let asset_lifecycles = self.server.asset_lifecycles.read();
+            for potential_free in potential_frees.drain(..) {
+                if let Some(&0) = ref_counts.get(&potential_free) {
+                    let type_uuid = match potential_free {
+                        HandleId::Id(type_uuid, _) => Some(type_uuid),
+                        HandleId::AssetPathId(id) => asset_sources
+                            .get(&id.source_path_id())
+                            .and_then(|source_info| source_info.get_asset_type(id.label_id())),
+                    };
+
+                    if let Some(type_uuid) = type_uuid {
+                        if let Some(asset_lifecycle) = asset_lifecycles.get(&type_uuid) {
+                            asset_lifecycle.free_asset(potential_free);
+                        }
+                    }
+                }
+            }
+        }
+    }
+
+    pub fn mark_unused_assets(&self) {
         let receiver = &self.server.asset_ref_counter.channel.receiver;
         let mut ref_counts = self.server.asset_ref_counter.ref_counts.write();
-        let asset_sources = self.server.asset_sources.read();
-        let mut potential_frees = Vec::new();
+        let mut potential_frees = None;
         loop {
             let ref_change = match receiver.try_recv() {
                 Ok(ref_change) => ref_change,
@@ -394,29 +460,11 @@ impl AssetServer {
                     let entry = ref_counts.entry(handle_id).or_insert(0);
                     *entry -= 1;
                     if *entry == 0 {
-                        potential_frees.push(handle_id);
-                    }
-                }
-            }
-        }
-
-        if !potential_frees.is_empty() {
-            let asset_lifecycles = self.server.asset_lifecycles.read();
-            for potential_free in potential_frees {
-                if let Some(i) = ref_counts.get(&potential_free).cloned() {
-                    if i == 0 {
-                        let type_uuid = match potential_free {
-                            HandleId::Id(type_uuid, _) => Some(type_uuid),
-                            HandleId::AssetPathId(id) => asset_sources
-                                .get(&id.source_path_id())
-                                .and_then(|source_info| source_info.get_asset_type(id.label_id())),
-                        };
-
-                        if let Some(type_uuid) = type_uuid {
-                            if let Some(asset_lifecycle) = asset_lifecycles.get(&type_uuid) {
-                                asset_lifecycle.free_asset(potential_free);
-                            }
-                        }
+                        potential_frees
+                            .get_or_insert_with(|| {
+                                self.server.asset_ref_counter.mark_unused_assets.lock()
+                            })
+                            .push(handle_id);
                     }
                 }
             }
@@ -432,7 +480,7 @@ impl AssetServer {
                 .expect("Asset should exist at this point.");
             if let Some(asset_lifecycle) = asset_lifecycles.get(&asset_value.type_uuid()) {
                 let asset_path =
-                    AssetPath::new_ref(&load_context.path, label.as_ref().map(|l| l.as_str()));
+                    AssetPath::new_ref(load_context.path, label.as_ref().map(|l| l.as_str()));
                 asset_lifecycle.create_asset(asset_path.into(), asset_value, load_context.version);
             } else {
                 panic!(
@@ -446,7 +494,9 @@ impl AssetServer {
         }
     }
 
-    pub(crate) fn update_asset_storage<T: Asset>(&self, assets: &mut Assets<T>) {
+    // Note: this takes a `ResMut<Assets<T>>` to ensure change detection does not get
+    // triggered unless the `Assets` collection is actually updated.
+    pub(crate) fn update_asset_storage<T: Asset>(&self, mut assets: ResMut<Assets<T>>) {
         let asset_lifecycles = self.server.asset_lifecycles.read();
         let asset_lifecycle = asset_lifecycles.get(&T::TYPE_UUID).unwrap();
         let mut asset_sources_guard = None;
@@ -471,7 +521,7 @@ impl AssetServer {
                         }
                     }
 
-                    let _ = assets.set(result.id, result.asset);
+                    assets.set_untracked(result.id, *result.asset);
                 }
                 Ok(AssetLifecycleEvent::Free(handle_id)) => {
                     if let HandleId::AssetPathId(id) = handle_id {
@@ -479,9 +529,7 @@ impl AssetServer {
                             .get_or_insert_with(|| self.server.asset_sources.write());
                         if let Some(source_info) = asset_sources.get_mut(&id.source_path_id()) {
                             source_info.committed_assets.remove(&id.label_id());
-                            if source_info.is_loaded() {
-                                source_info.load_state = LoadState::Loaded;
-                            }
+                            source_info.load_state = LoadState::Unloaded;
                         }
                     }
                     assets.remove(handle_id);
@@ -495,22 +543,35 @@ impl AssetServer {
     }
 }
 
-pub fn free_unused_assets_system(asset_server: Res<AssetServer>) {
+fn free_unused_assets_system_impl(asset_server: &AssetServer) {
     asset_server.free_unused_assets();
+    asset_server.mark_unused_assets();
+}
+
+pub fn free_unused_assets_system(asset_server: Res<AssetServer>) {
+    free_unused_assets_system_impl(&asset_server);
 }
 
 #[cfg(test)]
 mod test {
     use super::*;
+    use crate::{loader::LoadedAsset, update_asset_storage_system};
+    use bevy_ecs::prelude::*;
+    use bevy_reflect::TypeUuid;
     use bevy_utils::BoxedFuture;
 
+    #[derive(Debug, TypeUuid)]
+    #[uuid = "a5189b72-0572-4290-a2e0-96f73a491c44"]
+    struct PngAsset;
+
     struct FakePngLoader;
     impl AssetLoader for FakePngLoader {
         fn load<'a>(
             &'a self,
             _: &'a [u8],
-            _: &'a mut LoadContext,
+            ctx: &'a mut LoadContext,
         ) -> BoxedFuture<'a, Result<(), anyhow::Error>> {
+            ctx.set_default_asset(LoadedAsset::new(PngAsset));
             Box::pin(async move { Ok(()) })
         }
 
@@ -519,6 +580,21 @@ mod test {
         }
     }
 
+    struct FailingLoader;
+    impl AssetLoader for FailingLoader {
+        fn load<'a>(
+            &'a self,
+            _: &'a [u8],
+            _: &'a mut LoadContext,
+        ) -> BoxedFuture<'a, Result<(), anyhow::Error>> {
+            Box::pin(async { anyhow::bail!("failed") })
+        }
+
+        fn extensions(&self) -> &[&str] {
+            &["fail"]
+        }
+    }
+
     struct FakeMultipleDotLoader;
     impl AssetLoader for FakeMultipleDotLoader {
         fn load<'a>(
@@ -534,10 +610,10 @@ mod test {
         }
     }
 
-    fn setup() -> AssetServer {
+    fn setup(asset_path: impl AsRef<Path>) -> AssetServer {
         use crate::FileAssetIo;
 
-        let asset_server = AssetServer {
+        AssetServer {
             server: Arc::new(AssetServerInternal {
                 loaders: Default::default(),
                 extension_to_loader_index: Default::default(),
@@ -546,38 +622,39 @@ mod test {
                 handle_to_path: Default::default(),
                 asset_lifecycles: Default::default(),
                 task_pool: Default::default(),
-                asset_io: Box::new(FileAssetIo::new(&".")),
+                asset_io: Box::new(FileAssetIo::new(asset_path)),
             }),
-        };
-        asset_server.add_loader::<FakePngLoader>(FakePngLoader);
-        asset_server.add_loader::<FakeMultipleDotLoader>(FakeMultipleDotLoader);
-        asset_server
+        }
     }
 
     #[test]
     fn extensions() {
-        let asset_server = setup();
+        let asset_server = setup(".");
+        asset_server.add_loader(FakePngLoader);
+
         let t = asset_server.get_path_asset_loader("test.png");
         assert_eq!(t.unwrap().extensions()[0], "png");
     }
 
     #[test]
     fn case_insensitive_extensions() {
-        let asset_server = setup();
+        let asset_server = setup(".");
+        asset_server.add_loader(FakePngLoader);
+
         let t = asset_server.get_path_asset_loader("test.PNG");
         assert_eq!(t.unwrap().extensions()[0], "png");
     }
 
     #[test]
     fn no_loader() {
-        let asset_server = setup();
+        let asset_server = setup(".");
         let t = asset_server.get_path_asset_loader("test.pong");
         assert!(t.is_err());
     }
 
     #[test]
     fn multiple_extensions_no_loader() {
-        let asset_server = setup();
+        let asset_server = setup(".");
 
         assert!(
             match asset_server.get_path_asset_loader("test.v1.2.3.pong") {
@@ -612,15 +689,188 @@ mod test {
 
     #[test]
     fn filename_with_dots() {
-        let asset_server = setup();
+        let asset_server = setup(".");
+        asset_server.add_loader(FakePngLoader);
+
         let t = asset_server.get_path_asset_loader("test-v1.2.3.png");
         assert_eq!(t.unwrap().extensions()[0], "png");
     }
 
     #[test]
     fn multiple_extensions() {
-        let asset_server = setup();
+        let asset_server = setup(".");
+        asset_server.add_loader(FakeMultipleDotLoader);
+
         let t = asset_server.get_path_asset_loader("test.test.png");
         assert_eq!(t.unwrap().extensions()[0], "test.png");
     }
+
+    fn create_dir_and_file(file: impl AsRef<Path>) -> tempfile::TempDir {
+        let asset_dir = tempfile::tempdir().unwrap();
+        std::fs::write(asset_dir.path().join(file), &[]).unwrap();
+        asset_dir
+    }
+
+    #[test]
+    fn test_missing_loader() {
+        let dir = create_dir_and_file("file.not-a-real-extension");
+        let asset_server = setup(dir.path());
+
+        let path: AssetPath = "file.not-a-real-extension".into();
+        let handle = asset_server.get_handle_untyped(path.get_id());
+
+        let err = futures_lite::future::block_on(asset_server.load_async(path.clone(), true))
+            .unwrap_err();
+        assert!(match err {
+            AssetServerError::MissingAssetLoader { extensions } => {
+                extensions == ["not-a-real-extension"]
+            }
+            _ => false,
+        });
+
+        assert_eq!(asset_server.get_load_state(handle), LoadState::Failed);
+    }
+
+    #[test]
+    fn test_invalid_asset_path() {
+        let asset_server = setup(".");
+        asset_server.add_loader(FakePngLoader);
+
+        let path: AssetPath = "an/invalid/path.png".into();
+        let handle = asset_server.get_handle_untyped(path.get_id());
+
+        let err = futures_lite::future::block_on(asset_server.load_async(path.clone(), true))
+            .unwrap_err();
+        assert!(matches!(err, AssetServerError::AssetIoError(_)));
+
+        assert_eq!(asset_server.get_load_state(handle), LoadState::Failed);
+    }
+
+    #[test]
+    fn test_failing_loader() {
+        let dir = create_dir_and_file("fake.fail");
+        let asset_server = setup(dir.path());
+        asset_server.add_loader(FailingLoader);
+
+        let path: AssetPath = "fake.fail".into();
+        let handle = asset_server.get_handle_untyped(path.get_id());
+
+        let err = futures_lite::future::block_on(asset_server.load_async(path.clone(), true))
+            .unwrap_err();
+        assert!(matches!(err, AssetServerError::AssetLoaderError(_)));
+
+        assert_eq!(asset_server.get_load_state(handle), LoadState::Failed);
+    }
+
+    #[test]
+    fn test_asset_lifecycle() {
+        let dir = create_dir_and_file("fake.png");
+        let asset_server = setup(dir.path());
+        asset_server.add_loader(FakePngLoader);
+        let assets = asset_server.register_asset_type::<PngAsset>();
+
+        let mut world = World::new();
+        world.insert_resource(assets);
+        world.insert_resource(asset_server);
+
+        let mut tick = {
+            let mut free_unused_assets_system = free_unused_assets_system.system();
+            free_unused_assets_system.initialize(&mut world);
+            let mut update_asset_storage_system = update_asset_storage_system::<PngAsset>.system();
+            update_asset_storage_system.initialize(&mut world);
+
+            move |world: &mut World| {
+                free_unused_assets_system.run((), world);
+                update_asset_storage_system.run((), world);
+            }
+        };
+
+        fn load_asset(path: AssetPath, world: &World) -> HandleUntyped {
+            let asset_server = world.get_resource::<AssetServer>().unwrap();
+            let id = futures_lite::future::block_on(asset_server.load_async(path.clone(), true))
+                .unwrap();
+            asset_server.get_handle_untyped(id)
+        }
+
+        fn get_asset(id: impl Into<HandleId>, world: &World) -> Option<&PngAsset> {
+            world
+                .get_resource::<Assets<PngAsset>>()
+                .unwrap()
+                .get(id.into())
+        }
+
+        fn get_load_state(id: impl Into<HandleId>, world: &World) -> LoadState {
+            world
+                .get_resource::<AssetServer>()
+                .unwrap()
+                .get_load_state(id.into())
+        }
+
+        // ---
+        // Start of the actual lifecycle test
+        // ---
+
+        let path: AssetPath = "fake.png".into();
+        assert_eq!(LoadState::NotLoaded, get_load_state(path.get_id(), &world));
+
+        // load the asset
+        let handle = load_asset(path.clone(), &world);
+        let weak_handle = handle.clone_weak();
+
+        // asset is loading
+        assert_eq!(LoadState::Loading, get_load_state(&handle, &world));
+
+        tick(&mut world);
+        // asset should exist and be loaded at this point
+        assert_eq!(LoadState::Loaded, get_load_state(&handle, &world));
+        assert!(get_asset(&handle, &world).is_some());
+
+        // after dropping the handle, next call to `tick` will prepare the assets for removal.
+        drop(handle);
+        tick(&mut world);
+        assert_eq!(LoadState::Loaded, get_load_state(&weak_handle, &world));
+        assert!(get_asset(&weak_handle, &world).is_some());
+
+        // second call to tick will actually remove the asset.
+        tick(&mut world);
+        assert_eq!(LoadState::Unloaded, get_load_state(&weak_handle, &world));
+        assert!(get_asset(&weak_handle, &world).is_none());
+
+        // finally, reload the asset
+        let handle = load_asset(path.clone(), &world);
+        assert_eq!(LoadState::Loading, get_load_state(&handle, &world));
+        tick(&mut world);
+        assert_eq!(LoadState::Loaded, get_load_state(&handle, &world));
+        assert!(get_asset(&handle, &world).is_some());
+    }
+
+    #[test]
+    fn test_get_handle_path() {
+        const PATH: &str = "path/file.png";
+
+        // valid handle
+        let server = setup(".");
+        let handle = server.load_untyped(PATH);
+        let handle_path = server.get_handle_path(&handle).unwrap();
+
+        assert_eq!(handle_path.path(), Path::new(PATH));
+        assert!(handle_path.label().is_none());
+
+        let handle_id: HandleId = handle.into();
+        let path_id: HandleId = handle_path.get_id().into();
+        assert_eq!(handle_id, path_id);
+
+        // invalid handle (not loaded through server)
+        let mut assets = server.register_asset_type::<PngAsset>();
+        let handle = assets.add(PngAsset);
+        assert!(server.get_handle_path(&handle).is_none());
+
+        // invalid HandleId
+        let invalid_id = HandleId::new(Uuid::new_v4(), 42);
+        assert!(server.get_handle_path(invalid_id).is_none());
+
+        // invalid AssetPath
+        let invalid_path = AssetPath::new("some/path.ext".into(), None);
+        assert!(server.get_handle_path(invalid_path).is_none());
+    }
 }
